#!/usr/bin/env python
# -*- coding: utf-8 -*-

# |>
# | DATA FORMATS ANALYZER 1-1-1 - DFA CONVERTER PY [http://www.universalchemicalmarkup.org]
# | Updated: 2015-08-24
# | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
# | Author: Jan Mokrý (Main corresponding author:
# |                    Contact information is at the project website.)
# |>

import csv
import os
import re

import lxml.etree

import DfaOutput

# DFA CONVERTER PY - CONVERT SERVICE
class ConvertService(object):
    '''
    Provides methods to convert data from XML to DSV and back and also to TEX.

    **Parameters:**

    - verbosity: The *string* specifying verbosity. See details in :ref:`InitService`.
    - dsvDelimiter: The *string* specifying the delimiter for DSV files. Use comma (``','``) or tab (``'\\t'``) for compatibility.
    - refName: The *string* specifying the reference name. See details in :ref:`InitService`.
    - refVersion: The *string* specifying the reference version. See details in :ref:`InitService`.

    The values of following class variables can be adjusted in the source code to fit the needs of the specific analysis:
    *alreadyExplainedAbbreviations*, *outputFormatReviewed*, *outputFormatDescription* and *uppercaseLegend*.
    '''

    def __init__(self, verbosity='IWE', dsvDelimiter=',', refName=None, refVersion=None):
        '''
        Initializes class variables.
        '''
        self.outputService = DfaOutput.OutputService(verbosity=verbosity)
        self.refName = refName
        self.refVersion = refVersion
        self.dsvDelimiter = dsvDelimiter
        self.sourceFile = None
        self.outputFile = None
        self.outputTexFile = None
        # Adjust this to fit the needs of the specific analysis.
        self.alreadyExplainedAbbreviations = ['API', 'CDX', 'CDXML', 'CML', 'CTfile', 'DTD', 'InChI', 'Mol2', 'NCBI ASN.1',
                                              'NCBI XML', 'PDB', 'PDBx/mmCIF', 'PDBML', 'RELAX NG', 'SMILES', 'SLN', 'UCM', 'W3C',
                                              'XDfile', 'XHTML', 'XML', 'XSD']
        self.outputFormatReviewed = False
        self.outputFormatDescription = True
        self.uppercaseLegend = True

    def convertData(self, filepath=None):
        '''
        Converts data from XML to DSV and back and also to TEX. Conversion direction is determined by the current filetype
        extension. Conversion rules, are determined from the current file structure.

        **Parameters:**

        - filepath: The *string* specifying the full path to the file.
        '''
        filename, filetype = os.path.splitext(os.path.split(filepath)[1])
        filetype = filetype.lstrip('.').upper()

        try:
            if filetype == 'XML':
                self.sourceFile = lxml.etree.parse(filepath)

                if self.dsvDelimiter == ',': self.outputFile = file(filename + '.csv', 'w')
                elif self.dsvDelimiter == '\t': self.outputFile = file(filename + '.tsv', 'w')

                self.outputTexFile = file(filename.replace(self.refName + '--' + self.refVersion, self.refName.title()).replace(
                    self.refVersion, '').replace('--V-', '').replace('-', '') + '.tex', 'w')
                self.outputService.printMessage(message='INFO: Converting data - "XML -> DSV+TEX".')
                self._convertXmlToDsvAndTex()
                self.outputService.printMessage(message='INFO: All data converted.')
            elif (filetype == 'CSV') or (filetype == 'TSV'):
                self.sourceFile = file(filepath)
                self.outputFile = file(filename + '.xml', 'w')
                self.outputService.printMessage(message='INFO: Converting data - "DSV -> XML".')
                self._convertDsvToXml()
                self.outputService.printMessage(message='INFO: All data converted.')
            else:
                self.outputService.printMessage(
                    message='WARNING: Unable to determine conversion direction, file type not supported - "' + filetype + '"!')
                self.outputService.printMessage(message='INFO: Supported file types - "[\'XML\', \'CSV\', \'TSV\']".')
        except IOError:
            self.outputService.printMessage(message='WARNING: File not found - "' + filepath + '"!')

        if isinstance(self.sourceFile, file): self.sourceFile.close()

        if isinstance(self.outputFile, file): self.outputFile.close()

        if isinstance(self.outputTexFile, file): self.outputTexFile.close()

    def _convertXmlToDsvAndTex(self):
        '''
        Converts data from XML to DSV and TEX.
        '''
        columnsList = []
        rowsWriter = csv.writer(self.outputFile, delimiter=self.dsvDelimiter, lineterminator='\n')
        rowsContainer = self.sourceFile.getroot()
        rowContainerName = rowsContainer.tag.rstrip('s')
        rowsLegend = self._extractRowContainerLegend(rowsContainer.find(rowContainerName), '', []) # Extract legend strings.

        for rowCellLegend in rowsLegend:
            # From the XML structure extract data corresponding to columns in the resulting DSV file.
            if rowCellLegend.endswith(' (A)'):
                eName, aName = rowCellLegend.rsplit('/', 1)
                columnsList.append([rowContainer.get(aName[:-4]) for rowContainer in rowsContainer.findall('./' + eName)])
            elif rowCellLegend.endswith(' (D)'):
                tagPattern = re.compile('(\n)?</?(description|summary|detail|attributes|elements|types|software|'
                                        'functionality|modifiability|usability|performance|availability)/?>(\n)?')
                columnsList.append([tagPattern.sub(r'', lxml.etree.tostring(rowContainer, encoding='utf-8').strip())
                    for rowContainer in rowsContainer.findall('./' + rowCellLegend[:-4])])
            else:
                columnsList.append([rowContainer.text for rowContainer in rowsContainer.findall('./' + rowCellLegend)])

        # Write the top row with legend strings to the DSV output file.
        if self.uppercaseLegend: rowsWriter.writerow([rowCellLegend.replace('/', '-').upper() for rowCellLegend in rowsLegend])
        else: rowsWriter.writerow([rowCellLegend.replace('/', '-') for rowCellLegend in rowsLegend])

        rowCellDataToString = lambda rowCellData: rowCellData or '' # Replace the None object with an empty string.

        # Write rows with data strings to the DSV output file.
        for rowId in range(len(columnsList[0])):
            rowsWriter.writerow([rowCellDataToString(columnsList[columnId][rowId]).encode('utf-8')
                                 for columnId in range(len(columnsList))])

        # Write rows with data strings to the TEX output file.
        self.outputTexFile.write('% START OF CONTENT CONVERTED FROM XML DATA FILES.\n\n')
        urlMarkedPattern = re.compile('\\\\url{.*?}')
        urlPattern = re.compile('((http|https|ftp)://[-A-Za-z0-9\\.]+(:[0-9]*)?'
                                '[-_A-Za-z0-9\\$\\#\\%\\+\\*\\.\\!\\?\\(\\)\\[\\]\\{\\},;:@&=/~<>^`\\"\\\']*)')
        abbrPattern = re.compile('<abbr title="(.*?)">(.*?)</abbr>')
        aPattern = re.compile('<att>(.*?)</att>')
        ePattern = re.compile('<el>(.*?)</el>')
        processedAbbrStrings = self.alreadyExplainedAbbreviations

        for rowId in range(len(columnsList[0])):
            rowData = ''
            endDescription = False

            for columnId in range(len(columnsList)):
                rowCellLegend = rowsLegend[columnId]
                rowCellLegendNormCase = rowCellLegend.title()
                rowCellData = rowCellDataToString(columnsList[columnId][rowId]).encode('utf-8')

                # Skip empty row cells (elements or attributes in the source XML file).
                if rowCellData == '': continue

                # Skip row cells (elements or attributes in the source XML file), which shouldn't be
                # processed for the desired custom TEX output.
                if (((rowCellLegendNormCase == 'Format/Reviewed (A)') and not self.outputFormatReviewed)
                      or ((rowCellLegendNormCase == 'Format/Description (D)') and not self.outputFormatDescription)
                      or (rowCellLegendNormCase == 'Assessment/Referenced (A)')
                      or (rowCellLegendNormCase == 'Assessment/References (A)')
                      or (rowCellLegendNormCase == 'Attribute/Standard (A)')
                      or (rowCellLegendNormCase == 'Attribute/Concept')
                      or (rowCellLegendNormCase == 'Element/Standard (A)')
                      or (rowCellLegendNormCase == 'Element/Concept')):
                    continue

                if rowCellData == 'None': rowCellData = rowCellData.upper()
                else:
                    rowCellData = urlPattern.sub(r'\\url{\1}', rowCellData) # Markup detected links.
                    rowCellData = aPattern.sub(r'\\textit{\1}', rowCellData) # Markup detected attribute names.
                    rowCellData = ePattern.sub(r'\\textit{\1}', rowCellData) # Markup detected element names.
                    rowCellData = rowCellData.replace('%', '\\%').replace('_', '\\_') # Escape TEX special characters.

                    # Enhance the markup of detected links to be compatible with both hyperref and url TEX packages.
                    for urlString in re.findall(urlMarkedPattern, rowCellData):
                        rowCellData = rowCellData.replace(urlString, '\\ifthenelse{\\boolean{url}}{' + urlString + '}{'
                                                          + urlString.replace('\\%', '%').replace('\\_', '_') + '}')

                    # Handle abbreviations in the desired custom TEX output.
                    for abbrString in re.findall(abbrPattern, rowCellData):
                        replacedAbbrString = '<abbr title="' + abbrString[0] + '">' + abbrString[1] + '</abbr>'

                        if abbrString[1] not in processedAbbrStrings:
                            rowCellData = rowCellData.replace(replacedAbbrString, abbrString[1] + ' (' + abbrString[0] + ')', 1)
                            processedAbbrStrings.append(abbrString[1])
                        else:
                            rowCellData = rowCellData.replace(replacedAbbrString, abbrString[1])

                # Handle row cells (elements or attributes in the source XML file), which need
                # special processing for the desired custom TEX output.
                if rowCellLegendNormCase.endswith('Name (A)'):
                    if (rowCellLegendNormCase == 'Attribute/Name (A)'): rowCellData = 'Attribute ' + rowCellData
                    elif (rowCellLegendNormCase == 'Element/Name (A)'): rowCellData = 'Element ' + rowCellData

                    labelPattern = re.compile('\(.*?\)')
                    rowLabel = ' '.join(labelPattern.sub(r'', rowCellData).replace(', ', ' and ').translate(None, ',./').split())
                    rowData += ('\subsubsection*{' + rowCellData
                                + '}\label{subsubsec:' + rowLabel.replace(' ', '-').upper()
                                + '}\n\n<desc/>\\begin{description} \\itemsep0pt \\parskip0pt \\parsep0pt\n')
                    endDescription = True
                elif rowCellLegendNormCase.endswith(' (A)'):
                    rowData += '\\item[' + rowCellLegend.replace(rowContainerName + '/', '').replace(
                        ' (A)', '').replace('/', ' -- ').title() + ':] ' + rowCellData + '\n'
                elif rowCellLegendNormCase.endswith(' (D)'):
                    rowCellData = rowCellData.replace('<p>', '').replace('</p>', '\n').replace(
                        '<code>', '\\textit{').replace('</code>', '}').replace(
                        '<ul>', '\\begin{itemize}').replace('</ul>', '\\end{itemize}\n').replace(
                        '<li>', '\\item ').replace('</li>', '')

                    if rowCellLegendNormCase.endswith('Description (D)'):
                        rowData = rowData.replace('<desc/>', (rowCellData + '\n\n').replace('\n\n\n', '\n\n'))
                    elif ((rowCellLegendNormCase == 'Assessment/Detail (D)')
                          or (rowCellLegendNormCase == 'Concept/Attributes (D)')
                          or (rowCellLegendNormCase == 'Concept/Elements (D)')
                          or (rowCellLegendNormCase == 'Concept/Types (D)')
                          or (rowCellLegendNormCase == 'Format/Software (D)')):
                        rowData += (rowCellData.replace('\\item ', '\\item[' + rowCellLegend.replace(rowContainerName + '/',
                            '').replace(' (D)', '').replace('/', ' -- ').title() + ' (').replace(' -- ', '):] ') + '\n').replace(
                            '\n\n\n', '\n\n').replace(',', ', ').replace(',  ', ', ').replace(', ~', ',~')
                    else:
                        rowData += ('\\item['
                            + rowCellLegend.replace(rowContainerName + '/', '').replace(' (D)', '').replace('/', ' -- ').title()
                            + ':] ' + rowCellData + '\n').replace('\n\n\n', '\n\n').replace(',', ', ').replace(',  ', ', ').replace(
                            ', ~', ',~')
                elif rowCellLegendNormCase == 'Format/Website':
                    rowData += ('\\item[Website:] ' + rowCellData + '\n')
                elif rowCellLegendNormCase == 'Format/Links':
                    rowData += '\\item[Links:] ' + '\n\n'.join(rowCellData.split()) + '\n'
                # Handle row cells (elements or attributes in the source XML file), which need
                # just processing for the default TEX output.
                else:
                    rowData += ('\\item[' + rowCellLegend.replace(rowContainerName + '/', '').replace('/', ' -- ').title()
                                + ':] ' + rowCellData.replace(',', ', ').replace(',  ', ', ').replace(', ~', ',~') + '\n')

            if endDescription: rowData += '\\end{description}\n\n'

            rowData = rowData.replace('<desc/>', '').replace(
                '\\begin{description} \\itemsep0pt \\parskip0pt \\parsep0pt\n\\end{description}\n\n', '')
            self.outputTexFile.write(rowData)

        self.outputTexFile.write('% END OF CONTENT CONVERTED FROM XML DATA FILES.')

    def _convertDsvToXml(self):
        '''
        Converts data from DSV to XML.
        '''
        rowsReader = csv.reader(self.sourceFile, delimiter=self.dsvDelimiter, lineterminator='\n')
        rowsLegend = {}

        if self.uppercaseLegend: rawRowsLegend = [rowCellData.lower() for rowCellData in rowsReader.next()]
        else: rawRowsLegend = rowsReader.next()

        # Extract legend strings.
        for rowCellId, rowCellData in enumerate(rawRowsLegend): rowsLegend[rowCellId] = rowCellData.strip().split('-')[1:]

        # According to legend strings re-create the root element of the XML structure.
        rowContainerName = rawRowsLegend[0].strip().split('-', 1)[0]
        rowsContainer = lxml.etree.Element(rowContainerName + 's')
        rowsContainer.text = '\n\n'

        if (rowContainerName == 'attribute') or (rowContainerName == 'element') or (rowContainerName == 'type'):
            rowsContainer.set('mat', 'Material(\'Ref--' + self.refName + '--' + self.refVersion
                              + '\', \'science\', \'' + self.refName + ' ' + self.refVersion + ' Reference\')')

        # According to legend strings re-create the XML structure corresponding to rows in the processed DSV file.
        for rowData in rowsReader:
            rowContainer = lxml.etree.Element(rowContainerName)
            rowContainer.text = '\n'
            rowContainer.tail = '\n\n'
            rowLengthDifference = len(rowsLegend) - len(rowData)

            if rowLengthDifference:
                rowData += rowLengthDifference * ['']

            for rowCellId, rowCellData in enumerate(rowData):
                rowContainer = self._appendRowContainerContent(rowContainer, rowsLegend[rowCellId],
                    unicode(rowCellData.strip(), 'utf-8')) # Load data as unicode objects.

            rowsContainer.append(rowContainer)

        # Write the re-created XML structure to the XML output file.
        self.outputFile.write(lxml.etree.tostring(lxml.etree.ProcessingInstruction(
            'xml', text='version="1.0" encoding="utf-8"'), encoding='utf-8') + '\n\n')
        # The CDATA sections are removed before the output is written to the XML output file.
        self.outputFile.write(lxml.etree.tostring(
            rowsContainer, encoding='utf-8').replace('<![CDATA[', '').replace(']]>', ''))

    def _extractRowContainerLegend(self, rowContainer, rowCellLegend, rowsLegend):
        '''
        Extracts information about the processed XML structure. This information is then used as the legend to convert data
        from XML to DSV. Recursive approach makes it possible to handle more complicated XML tree structures.

        **Parameters:**

        - rowContainer: The *lxml.etree.Element* with the current container element with data corresponding
          to one or more row cell (depending on the recursion progress) in the resulting DSV file.
        - rowCellLegend: The *string* with the legend string in construction (depending on the recursion progress) corresponding
          to one row cell in the resulting DSV file.
        - rowsLegend: The *list* of legend strings corresponding to one or more row cell (depending on the recursion progress)
          in the resulting DSV file. It has the following structure: ``LIST[STRING]``.

        **Returns:** The *list* of legend strings corresponding to one or more row cell (depending on the recursion progress)
        in the resulting DSV file. It has the following structure: ``LIST[STRING]``.
        '''
        eName = rowContainer.tag

        if ((eName == 'description') or (eName == 'summary') or (eName == 'detail')
            or (eName == 'attributes') or (eName == 'elements') or (eName == 'types') or (eName == 'software')
            or (eName == 'functionality') or (eName == 'modifiability') or (eName == 'usability')
            or (eName == 'performance') or (eName == 'availability')):
            rowsLegend += [(rowCellLegend + '/' + eName + ' (D)').lstrip('/')]
        elif len(rowContainer) == 0: rowsLegend += [(rowCellLegend + '/' + eName).lstrip('/')] + [
            (rowCellLegend + '/' + eName + '/' + aName + ' (A)').lstrip('/')
            for aName in sorted(rowContainer.attrib.keys())]
        else:
            rowsLegend += [(rowCellLegend + '/' + eName + '/' + aName + ' (A)').lstrip('/')
                           for aName in sorted(rowContainer.attrib.keys())]

            for rowSubContainer in list(rowContainer):
                rowsLegend = self._extractRowContainerLegend(rowSubContainer, rowCellLegend + '/' + eName, rowsLegend)

        return rowsLegend

    def _appendRowContainerContent(self, rowContainer, rowCellLegend, rowCellData):
        '''
        Appends the content to the container element in order to re-create XML structure according to the legend during
        the conversion of data from DSV to XML. Recursive approach makes it possible to handle more complicated XML tree structures.

        **Parameters:**

        - rowContainer: The *lxml.etree.Element* with the current container element with data corresponding
          to one or more row cell (depending on the recursion progress) in the processed DSV file.
        - rowCellLegend: The *list* of legend strings to be processed (depending on the recursion progress) in order to re-create
          the XML structure for the data of one row cell from the processed DSV file. It has the following
          structure: ``LIST[STRING]``.
        - rowCellData: The *string* with the data of one row cell from the processed DSV file.

        **Returns:** The *lxml.etree.Element* with the current container element holding data corresponding to one or more
        row cell (depending on the recursion progress) in the processed DSV file.
        '''
        if len(rowCellLegend) == 1:
            rowCellLegend = rowCellLegend[0]

            if rowCellLegend.endswith((' (a)', ' (A)')): rowContainer.set(rowCellLegend[:-4], rowCellData)
            elif rowCellLegend.endswith((' (d)', ' (D)')):
                rowSubContainer = lxml.etree.Element(rowCellLegend[:-4])
                # The CDATA section is used to ensure that the text containing the description is not escaped,
                # so all markup inside the text will work.
                rowSubContainer.text = lxml.etree.CDATA('\n' + rowCellData + '\n')
                rowSubContainer.tail = '\n'
                rowContainer.append(rowSubContainer)
            else:
                rowSubContainer = None

                for rowSubContainerCandidate in list(rowContainer):
                    if rowSubContainerCandidate.tag == rowCellLegend:
                        rowSubContainer = rowSubContainerCandidate
                        rowSubContainer.text += ' ' + rowCellData
                        break

                if rowSubContainer is None:
                    rowSubContainer = lxml.etree.Element(rowCellLegend)
                    rowSubContainer.text = rowCellData
                    rowSubContainer.tail = '\n'
                    rowContainer.append(rowSubContainer)
        else:
            rowSubContainer = None

            for rowSubContainerCandidate in list(rowContainer):
                if rowSubContainerCandidate.tag == rowCellLegend[0]:
                    rowSubContainer = rowSubContainerCandidate
                    self._appendRowContainerContent(rowSubContainer, rowCellLegend[1:], rowCellData)
                    break

            if rowSubContainer is None:
                rowSubContainer = lxml.etree.Element(rowCellLegend[0])
                rowSubContainer.text = '\n'
                rowSubContainer.tail = '\n'
                rowContainer.append(self._appendRowContainerContent(rowSubContainer, rowCellLegend[1:], rowCellData))

        return rowContainer
